# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Remove RSE protocol default constraint"""    # noqa: D400, D415

from alembic import context
from alembic.op import alter_column

# Alembic revision identifiers
revision = '3b943000da18'
down_revision = '0548e199f091'

PRIORITY_COLUMNS = ['read_lan', 'write_lan', 'delete_lan',
                    'read_wan', 'write_wan', 'delete_wan',
                    'third_party_copy_read', 'third_party_copy_write']


def upgrade():
    """Upgrade the database to this revision."""
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        for column in PRIORITY_COLUMNS:
            alter_column('rse_protocols', column, server_default=None, schema=schema)


def downgrade():
    """Downgrade the database to the previous revision."""
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        for column in PRIORITY_COLUMNS:
            alter_column('rse_protocols', column, server_default='0', schema=schema)
