
'''
This file is automatically generated; Do not edit it. :)
'''
IS_FINAL = True
VERSION = '39.4.1'
BRANCH_NICK = 'release-39'
REVISION_ID = 'ec9c391d2816ac0b9fe0f45ed78bc8a05afcf8b2'
REVNO = '14337'
